/***************************************************************************
                           frame_memory.cpp  -  description
                             -------------------
    begin                : Mit Jun 25 2003
    copyright            : (C) 2003 by Dirk Henrici
    email                : henrici@informatik.uni-kl.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of license found in "COPYING".                     *
 *                                                                         *
 ***************************************************************************/

#include "frame_memory.h"

/**Konstruktor
  */
tAbstractMemoryPool::tAbstractMemoryPool(bool master, const string keystr, long memsize) : _master(master), _memsize(memsize) {
  if (keystr=="") {}
}

/**Konstruktor
  */
tSharedMemoryPool::tSharedMemoryPool(bool master, const string keystr, long memsize) : tAbstractMemoryPool(master, keystr, memsize), _shmid(0) {
  //Deskriptor-Key aus Key-String errechnen
  long key = crc32(keystr);
  if (keystr=="")
    debugprint('E',"key-string is missing", "tSharedMemoryPool::tSharedMemoryPool");
  //Sollte SharedMemory-Segment beim Master schon bestehen: freigeben  
  if (_master)
    while ((_shmid = shmget(key, memsize, 0)) >= 0)
      if (shmctl(_shmid, IPC_RMID, NULL)==-1)
        debugprintunix('E',"error destroying already exisiting shared memory", "tSharedMemoryPool::tSharedMemoryPool");   
  //Shared-Memory anlegen/ansprechen
  int attrib = (_master)?(IPC_CREAT | 0666):0;
  if ((_shmid = shmget(key, memsize, attrib)) < 0)
    debugprintunix('E',"error allocating shared memory", "tSharedMemoryPool::tSharedMemoryPool");
  debugprint('N', "shared memory id is: "+int2str(_shmid), "tSharedMemoryPool::tSharedMemoryPool");    
  //Shared-Memory referenzieren
  if ((_memptr = shmat(_shmid, NULL, 0)) == (char *) -1) {
    debugprintunix('E',"error referencing shared memory", "tSharedMemoryPool::tSharedMemoryPool");
    _memptr = NULL;    
  } else {  
    if (_master) {
      *(void**)_memptr = _memptr;
    } else {
      //ist die automatisch generierte Base-Adresse nicht die gleiche wie beim Master, Shared-Memory entsprechend neu referenzieren
      void* newmembase = *(void**)_memptr;
      if (newmembase != _memptr) {
        debugprint('I',"membase of master and client doesn't match, redefining client's base", "tSharedMemoryPool::tSharedMemoryPool");
        if (shmdt(_memptr) == -1)
          debugprintunix('E',"error detaching from shared memory", "tSharedMemoryPool::tSharedMemoryPool");
        if ((_memptr = shmat(_shmid, newmembase, 0)) == (char *) -1) {
          debugprintunix('E',"error referencing shared memory (base "+ptr2str(newmembase)+")", "tSharedMemoryPool::tSharedMemoryPool");
          _memptr = NULL;
        }  
      }
    }  
  }  
}

/**Destruktor
  */
tSharedMemoryPool::~tSharedMemoryPool() {
  if (_memptr!=NULL) {
    if (shmdt(_memptr) == -1)
      debugprintunix('E',"error detaching from shared memory", "tSharedMemoryPool::~tSharedMemoryPool");
    if (_master)  
      if (shmctl(_shmid, IPC_RMID, NULL)==-1)
        debugprintunix('E',"error destroying shared memory", "tSharedMemoryPool::~tSharedMemoryPool");
  }      
}

/**Konstruktor
  *Erwartet als Parameter einen Memory-Pool und einen Mutex deren Verwaltung komplett bernommen wird (inkl. Lschung)
  */
tMemoryAllocator::tMemoryAllocator(bool master, tAbstractMemoryPool* mempool, tAbstractMutex* mutex) : _master(master), _mempool(mempool), _mutex(mutex) {
  //Parameter auf Plausibilitt prfen, Checks
  if (mempool==NULL)
    debugprint('E',"no memory-pool for memory-allocator specified", "tMemoryAllocator::tMemoryAllocator");  
  if (mutex==NULL)
    debugprint('E',"no lock for memory-allocator specified", "tMemoryAllocator::tMemoryAllocator");
  tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)(_mempool->getmembase());
  debugprint('N', "memory-pool-header-address is: "+ptr2str(pMemPoolHeader), "tMemoryAllocator::tMemoryAllocator");      
  if (pMemPoolHeader!=NULL) {
    if (master)
      memcpy(pMemPoolHeader->identifier, MemPoolHeaderIdentifier, sizeof(MemPoolHeaderIdentifier));
    if (memcmp(pMemPoolHeader->identifier, MemPoolHeaderIdentifier, sizeof(MemPoolHeaderIdentifier))!=0)
      debugprint('E',"invalid mempool identifier", "tMemoryAllocator::tMemoryAllocator");      
    reset();
  }
}

/**Destruktor
  */
tMemoryAllocator::~tMemoryAllocator() {
  if (_mempool->getmembase()!=NULL)
    _mutex->remove();
  delete _mutex;
  delete _mempool;
}         
 
/**Liefert einen Info-String ber den angefragten Block zurck
  */
string tMemoryAllocator::getinfo(const tBlock block) {
  return getinfo(abs2offset(block.header));
}

/**Liefert einen Info-String ber den durch Zeiger auf Blockheader angefragten Block zurck
  */
string tMemoryAllocator::getinfo(offsettype block) {
  string info;
  tBlockHeader* header = offset2header(block);
  tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)(_mempool->getmembase());
  if (block==0) {
    info="Block is NULL";
  } else {  
    info="At "+int2str(block)+": ";
    info+="previous "+int2str(header->memoffset.previous)+" next "+int2str(header->memoffset.next)+": ";
    info+="previous key "+int2str(header->keyoffset.previous)+" next key "+int2str(header->keyoffset.next)+": ";        
    info+=" key value "+int2str(header->key) + " datasize "+int2str(header->datasize);  
    if (block == pMemPoolHeader->blocks.head)
      info=info+" head";
    if (block == pMemPoolHeader->blocks.tail)
      info=info+" tail";
    if (block == pMemPoolHeader->keys.head)
      info=info+" keys.head";
    if (block == pMemPoolHeader->keys.tail)
      info=info+" keys.tail";
    if (block == pMemPoolHeader->mem.start)
      info=info+" memstart";
    if (block+sizeof(block)+header->datasize == pMemPoolHeader->mem.end)
      info=info+" memfit";  
    if (block < pMemPoolHeader->mem.start)
      info=info+" BUFFER UNDERRUN";
    if (block+sizeof(block)+header->datasize > pMemPoolHeader->mem.end)
      info=info+" BUFFER OVERRUN";      
  }
  return info;  
}

/**Gibt zum Debuggen Informationen aus dem Header des Memory-Segmentes aus
  */
void tMemoryAllocator::dumpmeminfo() {
  tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)(_mempool->getmembase());
  string info=" blocks.head "+int2str(pMemPoolHeader->blocks.head)+" blocks.tail "+int2str(pMemPoolHeader->blocks.tail);
  info+=" keys.head "+int2str(pMemPoolHeader->keys.head)+" keys.tail "+int2str(pMemPoolHeader->keys.tail);  
  debugprint('I',info+": baseaddr "+ptr2str(pMemPoolHeader)+" blockmem ("+int2str(pMemPoolHeader->mem.start)+"="+ptr2str(offset2abs(pMemPoolHeader->mem.start))+" to "+int2str(pMemPoolHeader->mem.end)+"="+ptr2str(offset2abs(pMemPoolHeader->mem.end))+")", "tMemoryAllocator::dumpmeminfo");
}

/**Gibt zum Debuggen alle Blcke in der Reihenfolge, in der sie im Speicher liegen, aus
  */
void tMemoryAllocator::dumpblocks() {
  tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)(_mempool->getmembase());
  offsettype currentblock = pMemPoolHeader->blocks.head;
  int blockcount = 0;
  int totalsize = 0;
  debugprint('I',"MemoryAllocator-Memory-Chain: membaseaddr "+ptr2str(pMemPoolHeader)+" blocks ("+int2str(pMemPoolHeader->blocks.head)+"-"+int2str(pMemPoolHeader->blocks.tail)+"):", "tMemoryAllocator::dumpblocks");
  if (currentblock==0)
    debugprint('T',"Memory-Chain is empty");    
  while (currentblock!=0) {
    debugprint('T',getinfo(currentblock));
    blockcount++;
    totalsize+=offset2header(currentblock)->datasize;
    currentblock = offset2header(currentblock)->memoffset.next;
  }
  debugprint('T', "==> "+int2str(blockcount)+" blocks with "+int2str(totalsize)+" bytes of data.");
}

/**Gibt zum Debuggen alle Blcke anhand ihrer Schlssel aus
  */
void tMemoryAllocator::dumpkeys(const tKeyHint keyhint) {
  string info;  
  tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)(_mempool->getmembase());
  offsettype currentblock = pMemPoolHeader->keys.head;
  debugprint('I',"MemoryAllocator-Key-Chain: membaseaddr "+ptr2str(pMemPoolHeader)+" keys ("+int2str(pMemPoolHeader->keys.head)+"-"+int2str(pMemPoolHeader->keys.tail)+"):", "tMemoryAllocator::dumpkeys");
  if (currentblock==0)
    debugprint('T',"Key-Chain is empty");
  while (currentblock!=0) {
    tBlockHeader* header = offset2header(currentblock);  
    info="";
    if (header==keyhint.first())
      info+=" keyhint.first";
    if (header==keyhint.last())
      info+=" keyhint.last";
    debugprint('T',getinfo(currentblock)+info);
    currentblock = header->keyoffset.next;
  }
}

/**initialisiert den Memory-Pool-Header und den Mutex
  */
void tMemoryAllocator::reset() {
  char* memptr = _mempool->getmembase();
  offsettype offset = 0;
  tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)memptr;
  //Speicher im Memorypool in Header, Mutex und allokierbaren Speicher aufteilen
  offset+=sizeof(tMemPoolHeader);
  _mutex->connect(memptr+offset);
  offset+=_mutex->getmutexsize();
  _mutex->create();  
  //damit MemPool-Header initialisieren, wenn master
  if (_master) {
    pMemPoolHeader->mem.start = offset;
    pMemPoolHeader->mem.end = offset + _mempool->getmemsize();
    pMemPoolHeader->blocks.head = pMemPoolHeader->mem.end;  // ***+
    pMemPoolHeader->blocks.tail = 0;
    pMemPoolHeader->keys.head = 0;
    pMemPoolHeader->keys.tail = 0;
  }  
}

/**Sperrt den assoziierten Mutex
  */
void tMemoryAllocator::lock() {
  _mutex->acquire();
}

/**Entsperrt den assoziierten Mutex
  */
void tMemoryAllocator::unlock() {
  _mutex->release();
}

/**Allokiert einen Speicherblock angegebener Gre
  */
tMemoryAllocator::tBlock tMemoryAllocator::malloc(int size) {
  if (size==0)
    return tBlock(NULL);
  tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)(_mempool->getmembase());
  offsettype newblock = 0;
  offsettype currentblock = 0;
  tBlockHeader* newheader = NULL;
  tBlockHeader* currentheader = NULL;
  lock();
  //Mglichst weit vorne freien Speicherbereich suchen...
  if (pMemPoolHeader->mem.start+sizeof(tBlockHeader)+size<=pMemPoolHeader->blocks.head) {
    //freier Platz am Anfang vor erstem Block vorhanden bzw. erster Block?  
    newblock = pMemPoolHeader->mem.start;
  } else {
    //sonst weiter hinten bis zum Ende suchen  
    currentblock=pMemPoolHeader->blocks.head;
    while ((currentblock!=pMemPoolHeader->blocks.tail)&&(currentblock+2*sizeof(tBlockHeader)+offset2header(currentblock)->datasize+size>offset2header(currentblock)->memoffset.next))
      currentblock=offset2header(currentblock)->memoffset.next;
    if (currentblock+2*sizeof(tBlockHeader)+offset2header(currentblock)->datasize+size<=(offset2header(currentblock)->memoffset.next))
      newblock=currentblock+sizeof(tBlockHeader)+offset2header(currentblock)->datasize;
  }      
  newheader = offset2header(newblock);
  currentheader = offset2header(currentblock);
  //neuen Block verlinken
  if (newblock!=0) {
    if (currentblock==0) {
      newheader->memoffset.next = pMemPoolHeader->blocks.head;
      pMemPoolHeader->blocks.head = newblock;
    } else {
      newheader->memoffset.next = currentheader->memoffset.next;
      currentheader->memoffset.next = newblock;
    }
    if (newheader->memoffset.next!=pMemPoolHeader->mem.end) {
      offset2header(newheader->memoffset.next)->memoffset.previous = newblock;
    } else {
      pMemPoolHeader->blocks.tail = newblock;
    }
    newheader->memoffset.previous = currentblock;
  }
  if (newblock!=0) {
    newheader = offset2header(newblock);
    newheader->keyoffset.previous = 0;
    newheader->keyoffset.next = 0;
    newheader->key = 0;
    newheader->datasize = size;
  } else {
    debugprint('W',"no more free space in memory; have to return NULL-block", "tMemoryAllocator::malloc");
  }
  unlock();
  return tBlock(newheader);
}

/**Allokiert einen Speicherblock angegebender Gre und setzt Schlssel
  */
tMemoryAllocator::tBlock tMemoryAllocator::malloc_key(int size, long key) {
  tKeyHintStruc keyhint;
  keyhint.used = false;
  return malloc_keyhint(size, key, keyhint);
}

/**Allokiert einen Speicherblock angegebender Gre, setzt Schlssel und aktualisiert Hint zum schnellen Zugriff
  */
tMemoryAllocator::tBlock tMemoryAllocator::malloc_keyhint(int size, long key, tKeyHintStruc &keyhint) {
  lock(); 
  //Speicher allokieren
  tBlockHeader* newblock = malloc(size).header;
  if (newblock!=NULL) {
    offsettype newoffset = abs2offset(newblock);
    newblock->key = key;
    //Key-Kette aktualisieren
    if (key!=0) {
      tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)(_mempool->getmembase());    
      tBlockHeader* currentblock = end(key, tKeyHint(keyhint), false).current().header;
      //am Anfang einzufgender Block?
      if (currentblock==NULL) {
        if (pMemPoolHeader->keys.head==0) {
          pMemPoolHeader->keys.tail = newoffset;
        } else {
          offset2header(pMemPoolHeader->keys.head)->keyoffset.previous = newoffset;
          newblock->keyoffset.next = pMemPoolHeader->keys.head;
        }
        pMemPoolHeader->keys.head = newoffset;
        keyhint.first = newblock;
      } else {
        //sonst hinter angegebenem Block einfgen
        if (currentblock->key!=key)
          keyhint.first = newblock;
        newblock->keyoffset.next = currentblock->keyoffset.next;
        currentblock->keyoffset.next = newoffset;
        if (newblock->keyoffset.next!=0) {
          offset2header(newblock->keyoffset.next)->keyoffset.previous = newoffset;
        } else {
          pMemPoolHeader->keys.tail = newoffset;
        }
        newblock->keyoffset.previous = abs2offset(currentblock); 
      }
      keyhint.last = newblock;
      keyhint.used = true;
    }
  }  
  unlock();   
  return newblock;
}

/**Gibt einen Speicherblock frei
  */
void tMemoryAllocator::free(const tBlock &block) {
  lock();
  tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)(_mempool->getmembase());
  tBlockHeader* currentblock = block.header;
  if (currentblock!=NULL) {
    //Key-Kette aktualisieren
    //nchsten Block in Kette bearbeiten
    if (currentblock->keyoffset.next!=0) {
      offset2header(currentblock->keyoffset.next)->keyoffset.previous = currentblock->keyoffset.previous;
    } else {
      if (currentblock->key!=0) {
        pMemPoolHeader->keys.tail = currentblock->keyoffset.previous;
      }
    }
    //vorherigen Block in Kette bearbeiten
    if (currentblock->keyoffset.previous!=0) {
      offset2header(currentblock->keyoffset.previous)->keyoffset.next = currentblock->keyoffset.next;
    } else {
      if (currentblock->key!=0) {
        pMemPoolHeader->keys.head = currentblock->keyoffset.next;
      }
    }
    //Mem-Kette aktualisieren
    if (currentblock->memoffset.previous==0) {
      pMemPoolHeader->blocks.head = currentblock->memoffset.next;
    } else {
      offset2header(currentblock->memoffset.previous)->memoffset.next = currentblock->memoffset.next;
    }    
    if (currentblock->memoffset.next==pMemPoolHeader->mem.end) {
      //letzter Block? Tail anpassen
      pMemPoolHeader->blocks.tail = currentblock->memoffset.previous;
    } else {
      //sonst Zeiger auf vorherigen Block des nchsten Blocks anpassen
      offset2header(currentblock->memoffset.next)->memoffset.previous = currentblock->memoffset.previous;
    }
  } else {
    debugprint('W',"tried to free NULL-block", "tMemoryAllocator::free");
  }  
  unlock();
}

/**Gibt einen Speicherblock frei, aktualisiert Hint
  */
void tMemoryAllocator::free_keyhint(const tBlock &block, tKeyHintStruc &keyhint) {
  lock();
  //Hint aktualisieren
  tBlockHeader* currentblock = block.header;
  if (currentblock!=NULL) {
    //keyhint.last bearbeiten
    if (keyhint.last==currentblock) {
      keyhint.last = offset2header(currentblock->keyoffset.previous);
      if (keyhint.last!=0)
        if (keyhint.last->key!=currentblock->key)
          keyhint.last = 0;
    }  
    //keyhint.first bearbeiten
    if (keyhint.first==currentblock) {
      keyhint.first = offset2header(currentblock->keyoffset.next);
      if (keyhint.first!=0)      
        if (keyhint.first->key!=currentblock->key)
          keyhint.first = 0;
    }      
  }
  //Block ber normale Methode freigeben
  free(block);
  unlock();
}

/**Liefert den ersten Block mit "key" zurck, der Hint ermglicht optional direkten Zugriff
  *Ist "match" true, wird NULL zurckgeliefert, wenn das Element nicht gefunden wird, bei false das nchste Element.
  */
tMemoryAllocator::tBlock tMemoryAllocator::getfirst(const long key, const tKeyHint keyhint, bool match) {
  tBlockHeader* current = NULL;
  tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)(_mempool->getmembase());
  if (key!=0) {
    if (keyhint.used()) {
      //ggf. keyhint nutzen
      current = keyhint.first();
      if (current!=NULL)
        if (keyhint.first()->key!=key)
          debugprint('E',"keyhint doesn't match key", "tMemoryAllocator::begin");
    } else {
      //sonst mu gesucht werden...
      //vorne anfangen
      current = offset2header(pMemPoolHeader->keys.head);
      //alle Blcke entlang Keykette durchgehen bis Ende oder gefunden
      while ((current!=NULL)&&(current->key < key))
        current = offset2header(current->keyoffset.next);
      //wenn nicht gefunden, bei "match" Iterator nullen
      if (current!=NULL)
        if (match)
          if (current->key!=key)
            current = NULL;
    }
  } else {
    if (pMemPoolHeader->blocks.head != pMemPoolHeader->mem.end)    
      current = offset2header(pMemPoolHeader->blocks.head);
  }
  return current;
}

/**Liefert den letzten Block mit "key" zurck, der Hint ermglicht optional direkten Zugriff
  *Ist "match" true, wird NULL zurckgeliefert, wenn das Element nicht gefunden wird, bei false das nchste Element.
  */
tMemoryAllocator::tBlock tMemoryAllocator::getlast(const long key, const tKeyHint keyhint, bool match) {
  tBlockHeader* current = NULL;
  tMemPoolHeader* pMemPoolHeader = (tMemPoolHeader*)(_mempool->getmembase());
  if (key!=0) {
    if (keyhint.used()) {
      //ggf. keyhint nutzen
      current = keyhint.last();
      if (keyhint.last()->key!=key)
        debugprint('E',"keyhint doesn't match key", "tMemoryAllocator::end");
    } else {
      //sonst mu gesucht werden...
      //hinten anfangen
      current = offset2header(pMemPoolHeader->keys.tail);
      //alle Blcke entlang Keykette durchgehen bis Ende oder gefunden
      while ((current!=NULL)&&(current->key > key))
        current = offset2header(current->keyoffset.previous);
      //wenn nicht gefunden, bei "match" Iterator nullen
      if (current!=NULL)
        if (match)
          if (current->key!=key)
            current = NULL;
    }
  } else {
    current = offset2header(pMemPoolHeader->blocks.tail);
  }
  return current;
}

/**Liefert einen Iterator zurck, der auf das erste Element mit "key" verweist, der Hint ermglicht optional direkten Zugriff
  *Ist "match" true, wird NULL zurckgeliefert, wenn das Element nicht gefunden wird, bei false das nchste Element.
  */
tMemoryAllocator_Iterator tMemoryAllocator::begin(const long key, const tKeyHint keyhint, bool match) {
  return tMemoryAllocator_Iterator(this, key, getfirst(key, keyhint, match).header);
}

/**Liefert einen Iterator zurck, der auf das letzte Element mit "key" verweist, der Hint ermglicht optional direkten Zugriff
  *Ist "match" true, wird NULL zurckgeliefert, wenn das Element nicht gefunden wird, bei false das nchste Element.
  */
tMemoryAllocator_Iterator tMemoryAllocator::end(const long key, const tKeyHint keyhint, bool match) {
  return tMemoryAllocator_Iterator(this, key, getlast(key, keyhint, match).header);
}

/**Liefert die Daten des ersten Blocks mit "key" zurck
  */
void* tMemoryAllocator::getdata(const long key) {
  return getfirst(key).data;
}

/**Gibt den Block, auf den der Iterator gerade verweist, zurck
  */
tMemoryAllocator::tBlock tMemoryAllocator_Iterator::current() {
  return tMemoryAllocator::tBlock(_current);
}

/**Setzt Iterator auf den nachfolgenden Block, sofern vorhanden.
  *Ein key-aware Iterator verwendet die Key-Kette, sonst wird die MemBlock-Kette verwendet
  */
bool tMemoryAllocator_Iterator::next() {
  if (_current!=NULL) {
    if (_key!=0) {
      if (_current->keyoffset.next!=0) {
        if (_malloc->offset2header(_current->keyoffset.next)->key==_key) {
          _current = _malloc->offset2header(_current->keyoffset.next);
          return true;
        }    
      }     
    } else {
      tMemoryAllocator::tMemPoolHeader* pMemPoolHeader = (tMemoryAllocator::tMemPoolHeader*)(_malloc->_mempool->getmembase());    
      if (_current->memoffset.next!=pMemPoolHeader->blocks.head) {
        _current = _malloc->offset2header(_current->memoffset.next);
        return true;
      }    
    }
  }
  return false;
}

/**Setzt Iterator auf den vorherigen Block, sofern vorhanden.
  *Ein key-aware Iterator verwendet die Key-Kette, sonst wird die MemBlock-Kette verwendet
  */
bool tMemoryAllocator_Iterator::previous() {
  if (_current!=NULL) {
    if (_key!=0) {
      if (_current->keyoffset.previous!=0) {
        if (_malloc->offset2header(_current->keyoffset.previous)->key==_key) {
          _current = _malloc->offset2header(_current->keyoffset.previous);
          return true;
        }
      }
    } else {
      if (_current->memoffset.previous!=0) {
        _current = _malloc->offset2header(_current->memoffset.previous);
        return true;
      }
    }
  }
  return false;
}

