/***************************************************************************
                           main.cpp  -  description
                             -------------------
    begin                : Die Aug 5 2003
    copyright            : (C) 2003 by Dirk Henrici
    email                : henrici@informatik.uni-kl.de, news@henrici.biz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of license found in "COPYING".                     *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <cstdlib>
#include <map>
#include <set>
#include "frame_memory.h"

//!eindeutiger Identifier, wird ber CRC32 in eine ID umgerechnet
#define CFG_ShMemKey             "shared-memory example"
//!Gre des SharedMemory-Blocks
#define CFG_ShMemSize            4*512*1024
//!Identifier-String fr Semaphore, wird ber CRC32 in eine ID umgerechnet
#define CFG_ShMemSemName         "shared-memory example: semaphore"
//!Eindeutiger Key fr STL-Allocator
#define CFG_STLShMemAllocatorKey 5001

//!Jedes STL-Objekt im Shared Memory bekommt einen eindeutigen Key, mit dem wir von einem anderen Prozess aus das Objekt ansprechen knnen
#define CFG_KeyExampleMap        1
#define CFG_KeyExampleSet        2


/////////////////////////////////////////////////////////////////////////////
// master-process
/////////////////////////////////////////////////////////////////////////////
void main_master() {
  debugprint('I', "I'm the MASTER process", "MASTER");
  
  //initalize shared memory allocator
  debugprint('N', "creating memory allocator that uses a shared-memory-segment", "master");  
  bool master = true;
  tMemoryAllocator myalloc(master, new tSharedMemoryPool(master, CFG_ShMemKey, CFG_ShMemSize), new tSystemVSemaphore(master, CFG_ShMemSemName));
  
  //create STL-memory-allocator
  debugprint('N', "creating STL-allocator", "master");    
    //create short name for STL-Allocator with arbitrary type
  typedef tSTLShMemAllocator<bool> tSTLAllocator;
    //get memory block
  tMemoryAllocator::tBlock memblock = myalloc.malloc_key(sizeof(tSTLAllocator), CFG_STLShMemAllocatorKey);
    //create object with "placement new"
  tSTLAllocator* stl_allocator = new((memblock.data))tSTLAllocator;
    //register current process
  stl_allocator->setmalloc(&myalloc);

  //no all preparations are done
  //==> do something
  debugprint('N', "doing some examples", "master");      

  //MAP
    //create a map in shared memory
  debugprint('N', "creating 'map' in shared memory and adding some elements", "master");          
      //"typedef" so that we don't need to type that stuff more than once
  typedef map<int, int, std::less<int>, tSTLShMemAllocator<int> > tTestMap;
      //allocate memory for map
  memblock = myalloc.malloc_key(sizeof(tTestMap), CFG_KeyExampleMap);
      //initialize map-object using STL-allocator
  tTestMap* example_map = new((memblock.data))tTestMap(std::less<int>(), *stl_allocator);
  
    //Now the new map in shared memory can be used like an ordinary map
    //Notice that no other process may access simultaneously, probably a locking mechanism is required
    //Simplest form of locking (this locks the complete shared-memory-segment): "myalloc.lock();" and "myalloc.unlock();"
    //or "tGuardCurrentScope scopelock = tGuardCurrentScope(malloc.getmutex());" to lock within the current scope of variables
    
      //set some elements
  (*example_map)[5] = 1;
  (*example_map)[2] = 4;
  (*example_map)[1] = 3;
      //dump content
  cout << "example 'map' currently contains (master process):" << endl;
  tTestMap::iterator current = example_map->begin();
  while (current != example_map->end()) {
    cout << "  " << current->first << " ==> " << current->second << endl;
    current++;
  }
  
  //SET
    //create a set in shared memory
  debugprint('N', "creating 'set' in shared memory and adding some elements", "master");              
      //"typedef" so that we don't need to type that stuff more than once
  typedef set<int, std::less<int>, tSTLShMemAllocator<int> > tTestSet;
      //allocate memory for set    
  memblock = myalloc.malloc_key(sizeof(tTestSet), CFG_KeyExampleSet);
      //initialize set-object using STL-allocator  
  tTestSet* example_set = new((memblock.data))tTestSet(std::less<int>(), *stl_allocator);

    //Now the new set in shared memory can be used like an ordinary set
    //Notice that no other process may access simultaneously, probably a locking mechanism is required

      //set some elements
  example_set->insert(4);
  example_set->insert(42);
  example_set->insert(17);
      //dump content
  cout << "example 'set' currently contains (master process)" << endl;
  tTestSet::iterator current_setelem = example_set->begin();
  while (current_setelem != example_set->end()) {
    cout << "  " << *current_setelem << endl;
    current_setelem++;
  }      
  debugprint('I', "did some stuff\nnext process may do something\n", "master");      

  //wait for client to change our variables and do some STL stuff
  sleep(2);

  debugprint('I', "back in master process", "MASTER");        
  //dump content to show changes by other process(es)
  cout << "example 'map' currently contains (master process):" << endl;
  current = example_map->begin();
  while (current != example_map->end()) {
    cout << "  " << current->first << " ==> " << current->second << endl;
    current++;
  }
  cout << "example 'set' currently contains (master process)" << endl;
  current_setelem = example_set->begin();
  while (current_setelem != example_set->end()) {
    cout << "  " << *current_setelem << endl;
    current_setelem++;
  }
  
  //cleanup STL-objects
  debugprint('N', "cleaning up", "master");        
    //call destructor to deinitialize object and free memory block
  example_map->~tTestMap();
  myalloc.free(tMemoryAllocator::tBlock(NULL, example_map));
    //call destructor to deinitialize object and free memory block  
  example_set->~tTestSet();
  myalloc.free(tMemoryAllocator::tBlock(NULL, example_set));  

  //cleanup STL-allocator
  //This clears all process to memory-allocator mappings as the STL-allocator was the first object created
  stl_allocator->~tSTLAllocator();
  myalloc.free(tMemoryAllocator::tBlock(NULL, stl_allocator));    
            
  debugprint('I', "finished\n", "master");  
}


/////////////////////////////////////////////////////////////////////////////
// normal process (accesses shared memory segment created by master process)
/////////////////////////////////////////////////////////////////////////////
void main_normal() {
  //wait for master to initialize shared-memory
  sleep(1);
  debugprint('I', "I'm a NORMAL process", "NORMAL");  
  //attach to shared memory
  debugprint('N', "attaching to shared-memory-segment", "normal");  
  bool master = false;
  tMemoryAllocator myalloc(master, new tSharedMemoryPool(master, CFG_ShMemKey, CFG_ShMemSize), new tSystemVSemaphore(master, CFG_ShMemSemName));
  //initialize STL allocator
    //create short name for STL-Allocator-Type
  typedef tSTLShMemAllocator<bool> tSTLAllocator;
    //we can use the some object as the other process
  tSTLAllocator* stl_allocator = (tSTLAllocator*)myalloc.getdata(CFG_STLShMemAllocatorKey);
    //register current process
  stl_allocator->setmalloc(&myalloc);

  //no all preparations are done
  //==> look what others (here: only master) have done and make some changes

  //connect to example-map
    //same typedef as in master process
  typedef map<int, int, std::less<int>, tSTLShMemAllocator<int> > tTestMap;
    //get reference
  tTestMap* example_map = (tTestMap*)myalloc.getdata(CFG_KeyExampleMap);  

  //connect to example-set
    //same typedef as in master process
  typedef set<int, std::less<int>, tSTLShMemAllocator<int> > tTestSet;
    //get reference
  tTestSet* example_set = (tTestSet*)myalloc.getdata(CFG_KeyExampleSet);

  cout << "modifying example 'map' and example 'set' (normal process):" << endl;
 
  //add some elements
  cout << "adding 10 ==> 4" << endl;
  (*example_map)[10] = 4;
  cout << "adding 23" << endl;  
  example_set->insert(23);

  //delete some elements
  cout << "deleting 42" << endl;  
  example_set->erase(42);
  
  //dump content
  cout << "now example 'map' currently contains (normal process):" << endl;
  tTestMap::iterator current = example_map->begin();
  while (current != example_map->end()) {
    cout << "  " << current->first << " ==> " << current->second << endl;
    current++;
  }
  cout << "now example 'set' currently contains (normal process):" << endl;
  tTestSet::iterator current_setelem = example_set->begin();
  while (current_setelem != example_set->end()) {
    cout << "  " << *current_setelem << endl;
    current_setelem++;
  }
    
  debugprint('I', "finished\n", "normal");
}

int main(int argc, char *argv[])
{
  std::cout << "Running " << argv[0] << " as an example for using the STL in shared memory." << std::endl << std::endl;
  if (argc>1) std::cout << "no arguments required" << std::endl;

  //need to independent processes for our example
  if (fork()==0) {
    main_master();
  } else {
    main_normal();
  }  
  return EXIT_SUCCESS;
}
